/*
                      Joystick mapping structures

	Functions:

	int JSMapIsAllocated(int n)
	int JSMapAllocate()
	void JSMapSyncWithData(jsmap_struct *jsm)
	void JSMapDelete(int n)
	void JSMapDeleteAll()

	---


 */


#include "xsw.h"
#include "../include/jsw.h"


#ifdef JS_SUPPORT

int JSMapIsAllocated(int n)
{
	if((jsmap == NULL) ||
           (n < 0) ||
           (n >= total_jsmaps)
	)
	    return(0);
	else if(jsmap[n] == NULL)
	    return(0);
	else
	    return(1);
}


int JSMapAllocate()
{
	int i, n;
	jsmap_struct *jsmap_ptr;


	if(total_jsmaps < 0)
	    total_jsmaps = 0;


	for(i = 0; i < total_jsmaps; i++)
	{
	    if(jsmap[i] == NULL)
		break;
	}
	if(i < total_jsmaps)
	{
	    n = i;
	}
	else
	{
	    n = total_jsmaps;

	    total_jsmaps++;
	    jsmap = (jsmap_struct **)realloc(
		jsmap,
		total_jsmaps * sizeof(jsmap_struct *)
	    );
	    if(jsmap == NULL)
	    {
		total_jsmaps = 0;
		return(-1);
	    }

	    jsmap[n] = NULL;
	}

	/* Allocate structure as needed. */
	if(jsmap[n] == NULL)
	{
	    jsmap[n] = (jsmap_struct *)calloc(
		1,
		sizeof(jsmap_struct)
	    );
	    if(jsmap[n] == NULL)
	        return(-1);
	}


	jsmap_ptr = jsmap[n];


	/* Reset values. */
	jsmap_ptr->jsd.fd = -1;



	return(n);
}

void JSMapSyncWithData(jsmap_struct *jsm)
{
	int i, n;


	js_data_struct *jsd;


	if(jsm == NULL)
	    return;

	jsd = &jsm->jsd;

	if(jsd->fd < 0)
	{
	    fprintf(
		stderr,
		"JSMapSyncWithData(): Warning: Joystick data not initialized.\n"
	    );
	}


	/* *********************************************************** */
	/* Sanitize axis totals. */
	if(jsm->total_axises < 0)
	    jsm->total_axises = 0;
	if(jsd->total_axises < 0)
	    jsd->total_axises = 0;

	/* Resync number of axises. */
	if(jsm->total_axises == jsd->total_axises)
	{
	    /* Do nothing. */
	}
	else if(jsm->total_axises > jsd->total_axises)
	{
	    /* JSMap has too many axises. */

	    for(i = jsm->total_axises - 1; i >= jsd->total_axises; i--)
		free(jsm->axis[i]);

	    jsm->total_axises = jsd->total_axises;
	    if(jsm->total_axises > 0)
	    {
		jsm->axis = (jsmap_axis_struct **)realloc(
		    jsm->axis,
		    jsm->total_axises * sizeof(jsmap_axis_struct *)
		);
		if(jsm->axis == NULL)
		{
		    jsm->total_axises = 0;
		}
	    }
            else
            {
                free(jsm->axis);
                jsm->axis = NULL;  

                jsm->total_axises = 0;
            }
	}
	else if(jsm->total_axises < jsd->total_axises)
	{
	    /* JSMap needs more axises. */

	    /* Allocate new number of axises specififed by jsd. */
	    jsm->axis = (jsmap_axis_struct **)realloc(
                jsm->axis,
		jsd->total_axises * sizeof(jsmap_axis_struct *)
	    );
	    if(jsm->axis == NULL)
	    {
		jsm->total_axises = 0;
		return;
	    }

	    for(i = jsm->total_axises; i < jsd->total_axises; i++)
	    {
		jsm->axis[i] = (jsmap_axis_struct *)calloc(
		    1,
		    sizeof(jsmap_axis_struct)
		);
		if(jsm->axis[i] == NULL)
		    continue;

		/* Reset values. */
		jsm->axis[i]->op_code = JSMAP_AXIS_OP_NONE;

	    }

	    /* Set new total. */
	    jsm->total_axises = jsd->total_axises;
	}

        /* *********************************************************** */
        /* Sanitize button totals. */
        if(jsm->total_buttons < 0)
            jsm->total_buttons = 0;
        if(jsd->total_buttons < 0)
            jsd->total_buttons = 0;
                    
        /* Resync number of buttons. */
        if(jsm->total_buttons == jsd->total_buttons)
        {
            /* Do nothing. */
        }
        else if(jsm->total_buttons > jsd->total_buttons)
        {
            /* JSMap has too many buttons. */
        
            for(i = jsm->total_buttons - 1; i >= jsd->total_buttons; i--)
                free(jsm->button[i]);

            jsm->total_buttons = jsd->total_buttons;
            if(jsm->total_buttons > 0)
            {
                jsm->button = (jsmap_button_struct **)realloc(
                    jsm->button,
                    jsm->total_buttons * sizeof(jsmap_button_struct *)
                );
                if(jsm->button == NULL) 
                {
                    jsm->total_buttons = 0;
                }
            }
	    else
	    {
		free(jsm->button);
		jsm->button = NULL;

		jsm->total_buttons = 0;
	    }
        }
        else if(jsm->total_buttons < jsd->total_buttons)
        {
            /* JSMap needs more buttons. */

            /* Allocate new number of buttons specififed by jsd. */
            jsm->button = (jsmap_button_struct **)realloc(
                jsm->button,
                jsd->total_buttons * sizeof(jsmap_button_struct *)
            );  
            if(jsm->button == NULL)
            {       
                jsm->total_buttons = 0;
                return;
            }         
         
            for(i = jsm->total_buttons; i < jsd->total_buttons; i++)
            {
                jsm->button[i] = (jsmap_button_struct *)calloc(
                    1,
                    sizeof(jsmap_button_struct)
                );
                if(jsm->button[i] == NULL)
                    continue;
        
                /* Reset values. */
                jsm->button[i]->keycode = 0;
                jsm->button[i]->state = 0;
            }

            /* Set new total. */
            jsm->total_buttons = jsd->total_buttons;
        }


	return;
}


void JSMapDelete(int n)
{
	int i;


	if(JSMapIsAllocated(n))
	{
	    /* Free device name. */
	    free(jsmap[n]->device_name);

	    /* Close associated joystick as needed. */
	    JSClose(&jsmap[n]->jsd);

	    /* Free all axis mappings. */
	    for(i = 0; i < jsmap[n]->total_axises; i++)
	        free(jsmap[n]->axis[i]);
	    free(jsmap[n]->axis);

	    /* Free all button mappings. */
            for(i = 0; i < jsmap[n]->total_buttons; i++)
                free(jsmap[n]->button[i]);
            free(jsmap[n]->button);

	    free(jsmap[n]);
	    jsmap[n] = NULL;
	}


	return;
}


void JSMapDeleteAll()
{
	int i;


	for(i = 0; i < total_jsmaps; i++)
	    JSMapDelete(i);

	free(jsmap);
	jsmap = NULL;

	total_jsmaps = 0;


	return;
}


#endif	/* JS_SUPPORT */
