/*
                              Joystick wrapper
 */

#ifndef JOYSTICKW_H
#define JOYSTICKW_H

#ifdef JS_SUPPORT


/*
 *	OS Specific Includes and Definations:
 */
#ifdef __linux__
# include "linux/joystick.h"
# define JSDefaultDevice	"/dev/js0"
# define JSDefaultCaliberation	".joystick"
#endif

/* Assume defaults for device and caliberation file. */
#ifndef JSDefaultDevice
# define JSDefaultDevice		"/dev/js0"
#endif

#ifndef JSDefaultCaliberation
# define JSDefaultCaliberation		".joystick"
#endif



/* Default bounds. */
#ifndef JSDefaultMin
# define JSDefaultMin		0
#endif

#ifndef JSDefaultMax
# define JSDefaultMax		1000
#endif

#ifndef JSDefaultCenter
# define JSDefaultCenter	500
#endif

#ifndef JSDefaultNullZone
# define JSDefaultNullZone	100
#endif



/*
 *	Error codes:
 */
#define JSSuccess	0
#define JSError		1
#define JSBadValue	2
#define JSNoAccess	3
#define JSNoBuffers	4


/*
 *	Event codes:
 */
#define JSNoEvent	0
#define JSGotEvent	1

/*
 *	Button state codes:
 */
#define JSButtonStateOff	0
#define JSButtonStateOn		1


/*
 *	Joystick data structure
 */
typedef struct {

	/* All units in raw units. */
	int cur;		/* Current position. */
	int min, cen, max;	/* Bounds. */
	int nz;			/* Null zone, in raw units from cen. */

	char flip;		/* Flip values. */

} js_axis_struct;

typedef struct {

	unsigned char state;	/* One of JSButtonState*. */

} js_button_struct;

typedef struct {

	int x;			/* Not sure what to put in this one yet. */

} js_hat_struct;


typedef struct {

	/* Read only. */
	char *name;

	js_axis_struct **axis;
	int total_axises;

	js_button_struct **button;
	int total_buttons;

	js_hat_struct **hat;
	int total_hats;

	char *device_name;
	char *caliberation_file;

	time_t time;

	/* Private, do not access. */
	int fd;

} js_data_struct;


#ifdef __linux__
extern int JSLoadCaliberationLinux(js_data_struct *jsd);
#endif

extern int JSIsAxisAllocated(js_data_struct *jsd, int n);
extern int JSIsButtonAllocated(js_data_struct *jsd, int n);
extern int JSIsHatAllocated(js_data_struct *jsd, int n);

extern double JSGetAxisCoeff(js_data_struct *jsd, int n);
extern double JSGetAxisCoeffNZ(js_data_struct *jsd, int n);

extern int JSInit(
	js_data_struct *jsd,
	int argc, char *argv[]
);
extern int JSUpdate(js_data_struct *jsd);
extern void JSClose(js_data_struct *jsd);


#endif	/* JS_SUPPORT */

#endif	/* JOYSTICKW_H */
