#include "swserv.h"
  
 
/*
 *      Print, refresh, or reload items in memory
 *      (not counting processes).
 */
int CmdMemory(int condescriptor, char *arg)
{
	int i, status;
        char sndbuf[CS_DATA_MAX_LEN];
        long total_mem,
             aux_mem,  
             connection_mem,
             objects_mem,   
             recycle_buf_mem,
             opm_mem,
             ocs_mem,
             schedual_mem;
        long con_object_num;
	xsw_object_struct *con_obj_ptr;


        /* Get connection's object number (assumed valid). */
        con_object_num = connection[condescriptor]->object_num;
	con_obj_ptr = xsw_object[con_object_num];


        /* Check if allowed to view memory. */
        if(con_obj_ptr->permission.uid > ACCESS_UID_MEM)
        {
            sprintf(sndbuf,
   "memory: Access level %i: Permission denied.",
                ACCESS_UID_MEM
            );
            NetSendLiveMessage(condescriptor, sndbuf);
            return(-1);
        }


        /* Calculate memory being used. */
        SWServGetMemoryStats(
            &total_mem,
            &aux_mem,
            &connection_mem,
            &objects_mem,
            &recycle_buf_mem,
            &opm_mem,  
            &ocs_mem,  
            &schedual_mem   
        );
             
             
        /* Print memory stats. */
        if((arg == NULL) ? 1 : (arg[0] == '\0'))
        {
            sprintf(sndbuf,
"-        total       con      objs    recbuf      opms      ocss scheduals"
            );
            NetSendLiveMessage(condescriptor, sndbuf);

            sprintf(sndbuf,
"Mem:\
 %9ld\
 %9ld\
 %9ld\
 %9ld\
 %9ld\
 %9ld\
 %9ld",
                total_mem,
                connection_mem,
                objects_mem,
                recycle_buf_mem,
                opm_mem,
                ocs_mem,
                schedual_mem
            );
            NetSendLiveMessage(condescriptor, sndbuf);
            
            sprintf(sndbuf,
"Ent:\
 %9ld\
 %9d\
 %9ld\
 %9d\
 %9d\
 %9d\
 %9d",   
        
                total_objects + total_connections + total_recycled_objects +
                total_opms + total_ocss,
            
                total_connections,
                total_objects,
                total_recycled_objects,
                total_opms,
                total_ocss,
                total_scheduals
            );
            NetSendLiveMessage(condescriptor, sndbuf);
 

            return(0);
        }

 
        /* ********************************************************** */

        /* Reload files into memory? */
        if(!strcmp(arg, "reload"))
        {
            /* Permission check. */
            if(con_obj_ptr->permission.uid > ACCESS_UID_MEMOP)
            {
                sprintf(
		    sndbuf,
 "memory: Requires access level %i: Permission denied.",
                    ACCESS_UID_MEMOP
                );
                NetSendLiveMessage(condescriptor, sndbuf);
                return(-1);
            }


            NetSendLiveMessage(
		-1,
		"Server: Memory reload in progress..."
	    ); 

            sprintf(sndbuf,
                "Server: Reloading configuration from `%s'...",
                fname.rc
            );
            NetSendLiveMessage(condescriptor, sndbuf);
            RCLoadFromFile(fname.rc);
                
            sprintf(sndbuf,
                "Server: Reloading OPMs from `%s'...",
                fname.opm
            );
            NetSendLiveMessage(condescriptor, sndbuf);
            OPMLoadFromFile(fname.opm);
 
            sprintf(sndbuf,
                "Server: Reloading OCSs from `%s'...",
                fname.ocs
            );
            NetSendLiveMessage(condescriptor, sndbuf);
            OCSLoadFromFile(fname.ocs);
            
            NetSendLiveMessage(-1,
                "Server: Memory reload complete."
            );
        }
        /* Refresh objects in memory. */
        else if(!strcmp(arg, "refresh"))
        {
	    /* Permission check. */
            if(con_obj_ptr->permission.uid > ACCESS_UID_MEMOP)
            {
                sprintf(
                    sndbuf,
 "memory: Requires access level %i: Permission denied.",
                    ACCESS_UID_MEMOP
                );
                NetSendLiveMessage(condescriptor, sndbuf);
                return(-1);
            }


            sprintf(
		sndbuf,
                "Server: Reclaiming memory..."
            );
            NetSendLiveMessage(condescriptor, sndbuf);

            DBReclaim();
            OPMReclaim();
            OCSReclaim();
            SchedualReclaim();
	    ConReclaim();

            sprintf(sndbuf,
                "Server: Memory reclaim complete."
            );
            NetSendLiveMessage(condescriptor, sndbuf);
        }
        /* Reload input universe file. */
        else if(strpfx(arg, "reload_universe"))
        {
            /* Permission check. */
            if(con_obj_ptr->permission.uid > ACCESS_UID_MEMRELOADUNV)
            {
                sprintf(
                    sndbuf,   
 "memory: Requires access level %i: Permission denied.",
                    ACCESS_UID_MEMRELOADUNV
                );
                NetSendLiveMessage(condescriptor, sndbuf);
                return(-1);
            }

	    /*   Some important notes about connection and object
	     *   pointers being reallocated:
	     *
	     *   The call to this function orignates at some point from
	     *   NetManageRecv() which uses the connection pointers and
	     *   assumes that they don't change.   However if global
	     *   total_connections drops to 0, it will exit and not
	     *   perform any further access of the old pointer array
	     *   so we can get away with it.
	     *
	     *   The global object pointers are not used (atleast
	     *   not in a loop including a call to this function)
	     *   in any functions stemming to this call, so it shouldn't
	     *   be a problem.
	     */

	    /* Check for double check. */
	    if(strcmp(arg, "reload_universe=YES"))
	    {
		sprintf(sndbuf,
 "memory: reload_universe: Argument must be typed `reload_universe=YES'."
		);
                NetSendLiveMessage(condescriptor, sndbuf);
	    }
	    else
	    {
		/* Notify connection of reload and the need to relogin. */
		NetSendLiveMessage(
		    -1,
 "Universe has been reloaded, you will need to relogin."
		);
		/* Disconnect all connections. */
                for(i = 0; i < total_connections; i++)
                    NetCloseConnection(i);


		/* Delete old universe and load new universe. */
		status = DBReadFromFile(fname.db_in);

		/* Delete all connections. */
		ConDeleteAll();
	    }
        }
	/* Unsupported operation. */
	else
	{
	    sprintf(
		sndbuf,
		"memory: %s: Unsupported operation.",
		arg
	    );
	    NetSendLiveMessage(condescriptor, sndbuf);
	}


        return(0);
}

