#include <fnmatch.h>

#include "swserv.h"



#define THIS_CMD_NAME	"sysparm"

#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#define MAX(a,b)        ((a) > (b) ? (a) : (b))



/*
 *      Set or view system parameters.
 */
int CmdSysparm(int condescriptor, char *arg)
{
        char sndbuf[CS_DATA_MAX_LEN + UNV_TITLE_MAX];

        char parameter[CS_DATA_MAX_LEN];
        char value[CS_DATA_MAX_LEN];

        char *strptr;
        long object_num;
	xsw_object_struct *obj_ptr;
        long con_object_num;
	connection_struct *con_ptr;        


	/* Get connection pointer (assumed valid). */
        con_ptr = connection[condescriptor];

	/* Get connection's object number and pointer. */
	con_object_num = con_ptr->object_num;
	if(DBIsObjectGarbage(con_object_num))
	    return(-1);
	else
	    obj_ptr = xsw_object[con_object_num];


        /* Check if allowed to view or set system parameters. */
        if(obj_ptr->permission.uid > ACCESS_UID_SYSPARM)
        {
            sprintf(sndbuf,
                "%s: Requires access level %i: Permission denied.",
		THIS_CMD_NAME,
                ACCESS_UID_SYSPARM
            );
            NetSendLiveMessage(condescriptor, sndbuf);
            return(-1);
        }

	StringStripSpaces(arg);


        /* Parse argument. */
        strptr = strchr(arg, '=');
        if(strptr == NULL)
	{
	    strncpy(parameter, arg, CS_DATA_MAX_LEN);
	    parameter[CS_DATA_MAX_LEN - 1] = '\0';

	    *value = '\0';
	}
	else
	{
            strncpy(value, strptr + 1, CS_DATA_MAX_LEN);
            value[CS_DATA_MAX_LEN - 1] = '\0';
            StringStripSpaces(value);

            *strptr = '\0';
            strncpy(parameter, arg, CS_DATA_MAX_LEN);
            parameter[CS_DATA_MAX_LEN - 1] = '\0';   
            StringStripSpaces(parameter);
	}


        /* Print parameter if value is empty. */
        if(*value == '\0')
        {
	    if(*parameter == '\0')
	    {
		strncpy(parameter, "*", CS_DATA_MAX_LEN);
		parameter[CS_DATA_MAX_LEN - 1] = '\0';
	    }

            /* unv_title */
            if(!fnmatch(parameter, "unv_title", 0))
            {
                sprintf(sndbuf,
                    "unv_title = %s",
                    unv_head.title
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

	    /* convert_ru_to_au */
            if(!fnmatch(parameter, "convert_ru_to_au", 0))
            {
                sprintf(sndbuf,
                    "convert_ru_to_au = %.4lf",
                    sw_units.ru_to_au
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* int_system_check */
	    if(!fnmatch(parameter, "int_system_check", 0))
	    {
                sprintf(sndbuf,
                    "int_system_check = %ld ms",
                    sysparm.int_system_check
                );
                NetSendLiveMessage(condescriptor, sndbuf);
	    }

            /* int_new_connection_poll */
            if(!fnmatch(parameter, "int_new_connection_poll", 0))
            {
                sprintf(sndbuf,
                    "int_new_connection_poll = %ld ms",
                    sysparm.int_new_connection_poll
                );
                NetSendLiveMessage(condescriptor, sndbuf);
	    }

            /* int_object_values */
            if(!fnmatch(parameter, "int_object_values", 0))
	    {
                sprintf(sndbuf,
                    "int_object_values = %ld ms",
                    sysparm.int_object_values
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* int_weapon_values */
            if(!fnmatch(parameter, "int_weapon_values", 0))
	    {
                sprintf(sndbuf,
                    "int_weapon_values = %ld ms",
                    sysparm.int_weapon_values
                );
                NetSendLiveMessage(condescriptor, sndbuf);
	    }

            /* int_memory_clean */
            if(!fnmatch(parameter, "int_memory_clean", 0))
            {
                sprintf(sndbuf,
                    "int_memory_clean = %ld ms",
                    sysparm.int_memory_clean
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* int_aux_con_stats */
            if(!fnmatch(parameter, "int_aux_con_stats", 0))
            {
                sprintf(sndbuf,
                    "int_aux_con_stats = %ld ms",
                    sysparm.int_aux_con_stats
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* int_os_stats (in seconds) */
            if(!fnmatch(parameter, "int_os_stats", 0))
            {
                sprintf(sndbuf,
                    "int_os_stats = %s",
                    StringFormatTimePeriod(sysparm.int_os_stats)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
                sprintf(sndbuf,
                    "    Next: %s",
                    StringFormatTimePeriod(next.os_stats - cur_systime)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* int_db_save (in seconds) */
            if(!fnmatch(parameter, "int_db_save", 0))
            {
                sprintf(sndbuf,
                    "int_db_save = %s",
                    StringFormatTimePeriod(sysparm.int_db_save)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
                sprintf(sndbuf,
                    "    Next: %s",
                    StringFormatTimePeriod(next.db_save - cur_systime)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* int_stats_export (in seconds) */
            if(!fnmatch(parameter, "int_stats_export", 0))
            {
                sprintf(sndbuf,
                    "int_stats_export = %s",
                    StringFormatTimePeriod(sysparm.int_stats_export)
                );  
                NetSendLiveMessage(condescriptor, sndbuf);
                sprintf(sndbuf,
                    "    Next: %s",
                    StringFormatTimePeriod(next.stats_export - cur_systime)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* max_connections */
            if(!fnmatch(parameter, "max_connections", 0))
            {
                sprintf(sndbuf,
                    "max_connections = %i",
                    sysparm.max_connections
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* max_guests */
            if(!fnmatch(parameter, "max_guests", 0))
            {
                sprintf(sndbuf,
                    "max_guests = %i",
                    sysparm.max_guests
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* max_aux_connections */
            if(!fnmatch(parameter, "max_aux_connections", 0))
            {
                sprintf(sndbuf,
                    "max_aux_connections = %i",
                    sysparm.max_aux_connections
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* login_timeout */
            if(!fnmatch(parameter, "login_timeout", 0))
            {
                sprintf(sndbuf,
                    "login_timeout = %s",
                    StringFormatTimePeriod(sysparm.login_timeout)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* max_failed_logins */
            if(!fnmatch(parameter, "max_failed_logins", 0))
            {
                sprintf(sndbuf,
                    "max_failed_logins = %i",
                    sysparm.max_failed_logins
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* guest_login_name */
            if(!fnmatch(parameter, "guest_login_name", 0))
            {
                sprintf(sndbuf, 
                    "guest_login_name = %s",
                    sysparm.guest_login_name
                );
                NetSendLiveMessage(condescriptor, sndbuf);
	    }

            /* allow_guest */
            if(!fnmatch(parameter, "allow_guest", 0))
            {
                sprintf(sndbuf,
                    "allow_guest = %s",
                    (sysparm.allow_guest) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

	    /* con_notify */
            if(!fnmatch(parameter, "con_notify", 0))
            {
                sprintf(sndbuf,
                    "con_notify = %s",
                    (sysparm.con_notify) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* single_connection */
            if(!fnmatch(parameter, "single_connection", 0))
            {
                sprintf(sndbuf,
                    "single_connection = %s",
                    ((sysparm.single_connection) ? "yes" : "no")
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* cease_fire */
            if(!fnmatch(parameter, "cease_fire", 0))
            {
                sprintf(sndbuf,
                    "cease_fire = %s",
                    ((sysparm.cease_fire) ? "yes" : "no")
                );
                NetSendLiveMessage(condescriptor, sndbuf);
	    }

            /* hide_players */
            if(!fnmatch(parameter, "hide_players", 0))
            {
                sprintf(sndbuf,
                    "hide_players = %s",
                    ((sysparm.hide_players) ? "yes" : "no")
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

	    /* homes_destroyable */
            if(!fnmatch(parameter, "homes_destroyable", 0))
            {
	        sprintf(sndbuf, 
                    "homes_destroyable = %s",
                    (sysparm.homes_destroyable) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
	    }

            /* lost_found_owner */
            if(!fnmatch(parameter, "lost_found_owner", 0))
            {
                sprintf(sndbuf,
                    "lost_found_owner = %s",
                    DBGetFormalNameStr(sysparm.lost_found_owner)
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* log_general */
            if(!fnmatch(parameter, "log_general", 0))
            {
                sprintf(sndbuf,
                    "log_general = %s",
                    (sysparm.log_general) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* log_events */
            if(!fnmatch(parameter, "log_events", 0))
            {
                sprintf(sndbuf,
                    "log_events = %s",
                    (sysparm.log_events) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* log_net */
            if(!fnmatch(parameter, "log_net", 0))
            {
                sprintf(sndbuf,
                    "log_net = %s", 
                    (sysparm.log_net) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* log_errors */
            if(!fnmatch(parameter, "log_errors", 0))
            {
                sprintf(sndbuf,
                    "log_errors = %s",
                    (sysparm.log_errors) ? "yes" : "no"
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* mesg_wrong_login */
            if(!fnmatch(parameter, "mesg_wrong_login", 0))
            {
                sprintf(sndbuf,
                    "mesg_wrong_login = %s",
                    sysparm.mesg_wrong_login
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* mesg_no_guests */
            if(!fnmatch(parameter, "mesg_no_guests", 0))
            {
                sprintf(sndbuf,
                    "mesg_no_guests = %s",
                    sysparm.mesg_no_guests
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* mesg_welcome */   
            if(!fnmatch(parameter, "mesg_welcome", 0))
            {
                sprintf(sndbuf,
                    "mesg_welcome = %s",
                    sysparm.mesg_welcome
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

            /* mesg_leave */
            if(!fnmatch(parameter, "mesg_leave", 0))
            {
                sprintf(sndbuf,
                    "mesg_leave = %s",
                    sysparm.mesg_leave
                );
                NetSendLiveMessage(condescriptor, sndbuf);
            }

	    return(0);
        }


	/* ******************************************************* */
	/* Set parameter to value. */

        /* unv_title */
        if(!strcmp(parameter, "unv_title"))
        {
	    strncpy(
		unv_head.title,
		value,
		UNV_TITLE_MAX
	    );
	    unv_head.title[UNV_TITLE_MAX - 1] = '\0';

            sprintf(sndbuf,
                "unv_title = %s",
                unv_head.title
            );
        }
	/* convert_ru_to_au */
        else if(!strcmp(parameter, "convert_ru_to_au"))
	{
            sw_units.ru_to_au = atof(value);
            if(sw_units.ru_to_au <= 0) 
                sw_units.ru_to_au = 1;

            sprintf(sndbuf,
                "convert_ru_to_au = %.4lf", 
                sw_units.ru_to_au
            );

	    /* Update units to all connections. */
	    NetSendUnits(-1);
	}
        /* int_system_check */
        else if(!strcmp(parameter, "int_system_check"))
        {
            sysparm.int_system_check = atol(value);
            if(sysparm.int_system_check < 1)
                sysparm.int_system_check = 1;

            sprintf(sndbuf,
                "int_system_check = %ld ms",
                sysparm.int_system_check
            );
        }
        /* int_new_connection_poll */
        else if(!strcmp(parameter, "int_new_connection_poll"))
        {
            sysparm.int_new_connection_poll = atol(value);
            if(sysparm.int_new_connection_poll < 1)
                sysparm.int_new_connection_poll = 1;

            sprintf(sndbuf,
                "int_new_connection_poll = %ld ms",
                sysparm.int_new_connection_poll
            );
        }
        /* int_object_values */
        else if(!strcmp(parameter, "int_object_values"))
        {
            sysparm.int_object_values = atol(value);
            if(sysparm.int_object_values < MIN_OBJECT_UPDATE_INT)
                sysparm.int_object_values = MIN_OBJECT_UPDATE_INT;
                
            sprintf(sndbuf,
                "int_object_values = %ld ms",
                sysparm.int_object_values
            );
        }
        /* int_weapon_values */
        else if(!strcmp(parameter, "int_weapon_values"))
        {
            sysparm.int_weapon_values = atol(value);
            if(sysparm.int_weapon_values < MIN_OBJECT_UPDATE_INT)
                sysparm.int_weapon_values = MIN_OBJECT_UPDATE_INT;

            sprintf(sndbuf,
                "int_weapon_values = %ld ms",
                sysparm.int_weapon_values
            );
        }
        /* int_memory_clean */
        else if(!strcmp(parameter, "int_memory_clean"))
        {
            sysparm.int_memory_clean = atol(value);
            if(sysparm.int_memory_clean < 1000)
                sysparm.int_memory_clean = 1000;
                
            sprintf(sndbuf,
                "int_memory_clean = %ld ms",
                sysparm.int_memory_clean
            );
        }
	/* int_aux_con_stats */
        else if(!strcmp(parameter, "int_aux_con_stats"))
        {
            sysparm.int_aux_con_stats = atol(value);
            if(sysparm.int_aux_con_stats < 1000)
                sysparm.int_aux_con_stats = 1000;

            sprintf(sndbuf,
                "int_aux_con_stats = %ld ms",
                sysparm.int_aux_con_stats
            );
        }
        /* int_os_stats */
        else if(!strcmp(parameter, "int_os_stats"))
        {
            sysparm.int_os_stats = atol(value);  /* In seconds. */
            if(sysparm.int_os_stats < 1)
                sysparm.int_os_stats = 1;

            sprintf(sndbuf,
                "int_os_stats = %s",
                StringFormatTimePeriod(sysparm.int_os_stats)
            );
        }
        /* int_db_save */
        else if(!strcmp(parameter, "int_db_save"))
        {
            sysparm.int_db_save = atol(value);  /* In seconds. */
            if(sysparm.int_db_save < 60)
                sysparm.int_db_save = 60;

            sprintf(sndbuf,
                "int_db_save = %s",
                StringFormatTimePeriod(sysparm.int_db_save)
            );
        }
        /* int_stats_export */
        else if(!strcmp(parameter, "int_stats_export"))
        {
            sysparm.int_stats_export = atol(value);  /* In seconds. */
            if(sysparm.int_stats_export < 1)
                sysparm.int_stats_export = 1;

            sprintf(sndbuf,
                "int_stats_export = %s",
                StringFormatTimePeriod(sysparm.int_stats_export)
            );
        }
        /* max_connections */
        else if(!strcmp(parameter, "max_connections"))
        {
            sysparm.max_connections = atoi(value);
            if(sysparm.max_connections < 1)
                sysparm.max_connections = 1;

            sprintf(sndbuf,
                "max_connections = %i",
                sysparm.max_connections
            );
        }
        /* max_guests */
        else if(!strcmp(parameter, "max_guests"))
        {
            sysparm.max_guests = atoi(value);
            if(sysparm.max_guests < 0)
                sysparm.max_guests = 0;

            sprintf(sndbuf,
                "max_guests = %i",
                sysparm.max_guests
            );
        }
        /* max_aux_connections */
        else if(!strcmp(parameter, "max_aux_connections"))
        {
            sysparm.max_aux_connections = atoi(value);
            if(sysparm.max_aux_connections < 0)
                sysparm.max_aux_connections = 0;
                
            sprintf(sndbuf,
                "max_aux_connections = %i",
                sysparm.max_aux_connections
            );
        }
        /* login_timeout */
        else if(!strcmp(parameter, "login_timeout"))
        {
            sysparm.login_timeout = atol(value); 
                
            sprintf(sndbuf,
                "login_timeout = %s", 
                StringFormatTimePeriod(sysparm.login_timeout)
            );
        }
        /* max_failed_logins */   
        else if(!strcmp(parameter, "max_failed_logins"))
        {
            sysparm.allow_guest = atoi(value);

            sprintf(sndbuf,
                "max_failed_logins = %i",
                sysparm.max_failed_logins
            );
        }
        /* guest_login_name */
        else if(!strcmp(parameter, "guest_login_name"))
        {
            strncpy(
                sysparm.guest_login_name,
                value,
                XSW_OBJ_NAME_MAX
            );
            sysparm.guest_login_name[XSW_OBJ_NAME_MAX - 1] = '\0';

            sprintf(sndbuf,
                "guest_login_name = %s",
                sysparm.guest_login_name
            );
        }
        /* allow_guest */
        else if(!strcmp(parameter, "allow_guest"))
        {
            sysparm.allow_guest = StringIsYes(value);

            sprintf(sndbuf,
                "allow_guest = %s",
                (sysparm.allow_guest) ? "yes" : "no"
            );
        }
        /* con_notify */
        else if(!strcmp(parameter, "con_notify"))
        {
            sysparm.con_notify = StringIsYes(value);

            sprintf(sndbuf,
                "con_notify = %s",
                (sysparm.con_notify) ? "yes" : "no"
            );
        }
	/* single_connection */
	else if(!strcmp(parameter, "single_connection"))
        {
	    sysparm.single_connection = StringIsYes(value);

            sprintf(sndbuf,
                "single_connection = %s",
                (sysparm.single_connection) ? "yes" : "no"
            );
        }
        /* cease_fire */
        else if(!strcmp(parameter, "cease_fire"))
        {
            sysparm.cease_fire = StringIsYes(value);
            
            sprintf(sndbuf,
                "cease_fire = %s",
                (sysparm.cease_fire) ? "yes" : "no"
            );
        }
        /* hide_players */
	else if(!strcmp(parameter, "hide_players"))
        {
            sysparm.hide_players = StringIsYes(value);

            sprintf(sndbuf,
                "hide_players = %s",
                (sysparm.hide_players) ? "yes" : "no"
            );
        }
        /* homes_destroyable */
        else if(!strcmp(parameter, "homes_destroyable"))
        {
            sysparm.homes_destroyable = StringIsYes(value);

            sprintf(sndbuf,
                "homes_destroyable = %s",
                (sysparm.homes_destroyable) ? "yes" : "no"
            );
        }
        /* lost_found_owner */
        else if(!strcmp(parameter, "lost_found_owner"))
        {
            /* Match a player object. */
            object_num = MatchObjectByName(value, XSW_OBJ_TYPE_PLAYER);
            if(DBIsObjectGarbage(object_num))
            {
                sprintf(sndbuf,
                    "%s: No such player object.",
                    value
                );
            }
            else
            {
                sysparm.lost_found_owner = object_num;

                sprintf(sndbuf,
                    "lost_found_owner = %s",
                    DBGetFormalNameStr(sysparm.lost_found_owner)
                );
            }
        }
        /* log_general */
        else if(!strcmp(parameter, "log_general"))
        {
            sysparm.log_general = StringIsYes(value);
                
            sprintf(sndbuf,
                "log_general = %s",
                (sysparm.log_general) ? "yes" : "no"
            );
        }
        /* log_events */
        else if(!strcmp(parameter, "log_events"))
        {
            sysparm.log_events = StringIsYes(value);
            
            sprintf(sndbuf,
                "log_events = %s",
                (sysparm.log_events) ? "yes" : "no"
            );
        }
        /* log_net */
        else if(!strcmp(parameter, "log_net"))
        {
            sysparm.log_net = StringIsYes(value);

            sprintf(sndbuf,
                "log_net = %s",
                (sysparm.log_net) ? "yes" : "no"
            );
        }
        /* log_errors */
        else if(!strcmp(parameter, "log_errors"))
        {
            sysparm.log_errors = StringIsYes(value);
        
            sprintf(sndbuf,
                "log_errors = %s",
                (sysparm.log_errors) ? "yes" : "no"
            );
        }
	/* mesg_wrong_login */
	else if(!strcmp(parameter, "mesg_wrong_login"))
        {
	    strncpy(
		sysparm.mesg_wrong_login,
		value,
		CS_MESG_MAX
	    );
	    sysparm.mesg_wrong_login[CS_MESG_MAX - 1] = '\0';

            sprintf(sndbuf,
                "mesg_wrong_login = %s",
                sysparm.mesg_wrong_login
            );
        }
        /* mesg_no_guests */
        else if(!strcmp(parameter, "mesg_no_guests"))
        {
            strncpy(
                sysparm.mesg_no_guests,
                value,
                CS_MESG_MAX
            );
            sysparm.mesg_no_guests[CS_MESG_MAX - 1] = '\0';

            sprintf(sndbuf,
                "mesg_no_guests = %s",
                sysparm.mesg_no_guests
            );
        }
        /* mesg_welcome */
        else if(!strcmp(parameter, "mesg_welcome"))
        {
            strncpy(
                sysparm.mesg_welcome,
                value,
                CS_MESG_MAX
            );
            sysparm.mesg_welcome[CS_MESG_MAX - 1] = '\0';

            sprintf(sndbuf,
                "mesg_welcome = %s",
                sysparm.mesg_welcome
            );
        }
        /* mesg_leave */
        else if(!strcmp(parameter, "mesg_leave"))
        {
            strncpy(
                sysparm.mesg_leave,
                value,
                CS_MESG_MAX
            );
            sysparm.mesg_leave[CS_MESG_MAX - 1] = '\0';

            sprintf(sndbuf,
                "mesg_leave = %s",
                sysparm.mesg_leave
            );
        }


        /* Unknown parameter. */
        else
        {
            sprintf(sndbuf,
                "%s: Unsupported sysparm.",
                parameter
            );
        }
        NetSendLiveMessage(condescriptor, sndbuf);


        return(0);
}
