/*
                                 File IO

	Functions:

	char *FReadNextLineAlloc(FILE *fp, char comment_char)

	---


*/

#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "../include/fio.h"


/*
 *	Returns an allocated string containing the entire
 *	line or NULL on error or EOF.
 *
 *	If comment_char is '\0' then the next line is read regardless
 *	if it is a comment or not.
 *
 *	Calling function must free() the returned pointer.
 */
char *FReadNextLineAlloc(FILE *fp, char comment_char)
{
	int i, m, n;
	char *strptr;


	if(fp == NULL)
	    return(NULL);

	/* Is comment character specified? */
	if(comment_char != '\0')
	{
	    /* Comment character is specified. */

	    /* Read past spaces, newlines, and comments. */
	    i = fgetc(fp);
	    if(i == EOF)
		return(NULL);

	    while((i == ' ') || (i == '\t') || (i == '\n') || (i == '\r') ||
                  (i == comment_char)
	    )
	    {
                if(i == EOF)
		    return(NULL);

		/* If comment, skip to next line. */
		if(i == comment_char)
		{
		    i = fgetc(fp);
		    while((i != '\n') && (i != '\r'))
		    {
			if(i == EOF)
			    return(NULL);
			i = fgetc(fp);
		    }
		}

		/* Get next character. */
		i = fgetc(fp);
	    }

	    /* Begin adding characters to string. */
	    m = 0;	/* mem size. */
            n = 1;	/* chars read. */
            strptr = NULL;

	    while((i != '\n') && (i != '\r') && (i != '\0'))
	    {
		/* Skip escaped newlines. */
		if(i == '\\')
		{
		    i = fgetc(fp);
		    if((i == '\n') || (i == '\r'))
		    {
			i = fgetc(fp);
		    }
		}

		if(i == EOF)
		    break;

		/* Allocate more memory as needed. */
		if(m < n)
		{
		    m += 5;	/* Allocate mem in 5 byte chunks. */
		    strptr = (char *)realloc(strptr, m * sizeof(char));
		    if(strptr == NULL) return(NULL);
		}

		strptr[n - 1] = (char)i;

                /* Read next character from file. */
		i = fgetc(fp);
		n++;	/* Increment characters read. */
	    }

	    /* NULL Terminate. */
	    m += 2;
            strptr = (char *)realloc(strptr, m * sizeof(char));
	    strptr[n - 1] = '\n';
	    strptr[n] = '\0';
	}
	else
	{
	    /* Comment character is not specified. */

            i = fgetc(fp);
            if(i == EOF)
		return(NULL);

            /* Begin adding characters to string. */
            m = 0;      /* Memory size. */
            n = 1;      /* Characters read. */
            strptr = NULL;	/* Return string. */
        
            while((i != '\n') && (i != '\r') && (i != EOF))
            {
                /* Allocate more memory as needed. */
                if(m < n)
                {
                    m += 5;	/* Allocate mem in 5 byte chunks. */
                    strptr = (char *)realloc(strptr, m * sizeof(char));
                    if(strptr == NULL)
			return(NULL);
                }
        
                strptr[n - 1] = (char)i;

		/* Read next character from file. */
                i = fgetc(fp);
                n++;	/* Increment characters read. */
            }

            /* NULL Terminate. */
            m += 2;
            strptr = (char *)realloc(strptr, m * sizeof(char));
            strptr[n - 1] = '\n';
            strptr[n] = '\0';
	}


	return(strptr);
}
