/*
                    Restart Configuration File

	Functions:

	int RCLoadFromFile(char *filename)


 */

#include "swserv.h"
#include "siteban.h"


int RCLoadFromFile(char *filename)
{
	int i, n, status;
        char *strptr, *strptr2, *strptr3;
        FILE *fp;
        off_t filesize;
        struct stat stat_buf;

        char parm[CFG_PARAMETER_MAX];
        char val[CFG_VALUE_MAX];

	char temp_name[NAME_MAX];

	int lines_read = 0;

	siteban_struct *sb_ptr;


        /* Check if exists. */
        if(stat(filename, &stat_buf))
        {
	    fprintf(stderr, "%s: No such file.\n", filename);
            return(-1);
        }

	/* Get size of file. */
        filesize = stat_buf.st_size;

        /* Open filename. */
        fp = fopen(filename, "r");
        if(fp == NULL)
        {
            fprintf(stderr, "%s: Cannot open file.\n",
                filename
            );
            return(-1);
        }


	/* *********************************************************** */

        strptr = NULL;

        while(1)
        {
            /* Free previous line and allocate/read next line. */
            free(strptr); strptr = NULL;
            strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
            if(strptr == NULL) break;
            lines_read++;

            /* Fetch parameter. */
            strptr2 = StringCfgParseParm(strptr);
            if(strptr2 == NULL) continue;
            strncpy(parm, strptr2, CFG_PARAMETER_MAX);
            parm[CFG_PARAMETER_MAX - 1] = '\0';

            /* Fetch value. */
            strptr2 = StringCfgParseValue(strptr);
            if(strptr2 == NULL) strptr2 = "0";  /* Set it to "0" if NULL. */
            strncpy(val, strptr2, CFG_VALUE_MAX);
            val[CFG_VALUE_MAX - 1] = '\0';


	    /* ListeningPort */
	    if(!strcasecmp(parm, "ListeningPort"))
	    {
		status = IncomingSocketInit(
		    atoi(val),
		    INCOMING_SOCKET_TYPE_STANDARD
		);
            }
            /* AUXStatsListeningPort */
            else if(!strcasecmp(parm, "AUXStatsListeningPort"))
            {
                status = IncomingSocketInit(
                    atoi(val),
                    INCOMING_SOCKET_TYPE_AUXSTATS
                );
            }
	    /* ServerToplevelDir */
	    else if(!strcasecmp(parm, "ServerToplevelDir"))
            {
#ifndef __WIN32__ /* Disable this for windows. */
		strncpy(dname.swserv_home_dir, val, PATH_MAX);
		dname.swserv_home_dir[PATH_MAX - 1] = '\0';

		/* Warn if not an absolute path. */
		if(!ISPATHABSOLUTE(dname.swserv_home_dir))
		{
                    fprintf(stderr,
                  "%s: Line %i: Warning: %s: Not absolute path statement.\n",
                        filename,  
                        lines_read,
                        dname.swserv_home_dir
                    );
		}

		/* Check if exists. */
		if(stat(dname.swserv_home_dir, &stat_buf))
		{
		    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
		        filename,
                        lines_read,
                        dname.swserv_home_dir
		    );
		}
#endif /* __WIN32__ */
            }
            /* ServerEtcDir */
            else if(!strcasecmp(parm, "ServerEtcDir"))
            {
                strncpy(dname.swserv_etc_dir, val, PATH_MAX);
                dname.swserv_etc_dir[PATH_MAX - 1] = '\0';
                        
                if(!ISPATHABSOLUTE(dname.swserv_etc_dir))
                {
                    strptr3 = PrefixPaths(
                        dname.swserv_home_dir,
                        dname.swserv_etc_dir
                    );
                    if(strptr3 != NULL)
                        strncpy(dname.swserv_etc_dir, strptr3, PATH_MAX);
                    dname.swserv_etc_dir[PATH_MAX - 1] = '\0';
                }

		/* Check if exists. */
                if(stat(dname.swserv_etc_dir, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
                        filename,
                        lines_read,
                        dname.swserv_etc_dir
                    );
                }
            }
            /* ServerLogsDir */
            else if(!strcasecmp(parm, "ServerLogsDir"))
            {
                strncpy(dname.swserv_logs_dir, val, PATH_MAX);
		dname.swserv_logs_dir[PATH_MAX - 1] = '\0';
                        
                if(!ISPATHABSOLUTE(dname.swserv_logs_dir))
                {
                    strptr3 = PrefixPaths(
                        dname.swserv_home_dir,
                        dname.swserv_logs_dir
                    );
                    if(strptr3 != NULL)
                        strncpy(dname.swserv_logs_dir, strptr3, PATH_MAX);
                    dname.swserv_logs_dir[PATH_MAX - 1] = '\0';
                }

		/* Check if exists. */
                if(stat(dname.swserv_logs_dir, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
                        filename,
                        lines_read,
                        dname.swserv_logs_dir
                    );   
                }
            }
            /* ServerBinDir */
            else if(!strcasecmp(parm, "ServerBinDir"))
            {
#ifndef __WIN32__ /* Disable item for windows. */
                strncpy(dname.swserv_bin_dir, val, PATH_MAX);
		dname.swserv_bin_dir[PATH_MAX - 1] = '\0';

                if(!ISPATHABSOLUTE(dname.swserv_bin_dir))
                {
                    strptr3 = PrefixPaths(
                        dname.swserv_home_dir,
                        dname.swserv_bin_dir
                    );
                    if(strptr3 != NULL)
                        strncpy(dname.swserv_bin_dir, strptr3, PATH_MAX);
                    dname.swserv_bin_dir[PATH_MAX - 1] = '\0';   
                }

		/* Check if exists. */
                if(stat(dname.swserv_bin_dir, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
                        filename,
                        lines_read,
                        dname.swserv_bin_dir
                    );
                }
#endif /* __WIN32__ */
            }
            /* ServerDBDir */
            else if(!strcasecmp(parm, "ServerDBDir"))
            {

                strncpy(dname.swserv_db_dir, val, PATH_MAX);
		dname.swserv_db_dir[PATH_MAX - 1] = '\0';

                if(!ISPATHABSOLUTE(dname.swserv_db_dir))
                {
                    strptr3 = PrefixPaths(
                        dname.swserv_home_dir,
                        dname.swserv_db_dir
                    );
                    if(strptr3 != NULL)
                        strncpy(dname.swserv_db_dir, strptr3, PATH_MAX);
                    dname.swserv_db_dir[PATH_MAX - 1] = '\0';
                }

		/* Check if exists. */
                if(stat(dname.swserv_db_dir, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
                        filename,
                        lines_read,
                        dname.swserv_db_dir
                    );
                }
            }
            /* ServerTmpDir */
            else if(!strcasecmp(parm, "ServerTmpDir"))
            {
		strncpy(dname.swserv_tmp_dir, val, PATH_MAX);
		dname.swserv_tmp_dir[PATH_MAX - 1] = '\0';

                if(!ISPATHABSOLUTE(dname.swserv_tmp_dir))
                {
		    strptr3 = PrefixPaths(
			dname.swserv_home_dir,
			dname.swserv_tmp_dir
		    );
		    if(strptr3 != NULL)
                        strncpy(dname.swserv_tmp_dir, strptr3, PATH_MAX);
		    dname.swserv_tmp_dir[PATH_MAX - 1] = '\0';
                }

		/* Check if exists. */
                if(stat(dname.swserv_tmp_dir, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
                        filename,
                        lines_read,
                        dname.swserv_tmp_dir
                    );
                }
            }
            /* PublicHTMLDir */
            else if(!strcasecmp(parm, "PublicHTMLDir"))
            {
		strncpy(dname.public_html_dir, val, PATH_MAX);
		dname.public_html_dir[PATH_MAX - 1] = '\0';

                if(!ISPATHABSOLUTE(dname.public_html_dir))
                {
		    strptr3 = PrefixPaths(
			dname.swserv_home_dir,
			dname.public_html_dir
		    );
		    if(strptr3 != NULL)
                        strncpy(dname.public_html_dir, strptr3, PATH_MAX);
		    dname.public_html_dir[PATH_MAX - 1] = '\0';
                }

		/* Check if exists. */
                if(stat(dname.public_html_dir, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such directory.\n",
                        filename,
                        lines_read,
                        dname.public_html_dir
                    );
                }
            }

	    /* ******************************************************** */
	    /* Monitor */
            else if(!strcasecmp(parm, "Monitor"))
            {
                if(fname.monitor_set == 0)
                {
                    strncpy(
                        fname.monitor,
                        val,
                        PATH_MAX + NAME_MAX
                    );
		    fname.monitor[PATH_MAX + NAME_MAX - 1] = '\0';
                    fname.monitor_set = 1;
                }

                if(!ISPATHABSOLUTE(fname.monitor))
                {
		    strptr3 = PrefixPaths(
                        dname.swserv_bin_dir,
                        fname.monitor
                    );
		    if(strptr3 != NULL)
                        strncpy(fname.monitor, strptr3, PATH_MAX + NAME_MAX);
		    fname.monitor[PATH_MAX + NAME_MAX - 1] = '\0';
                }

		/* Check if exists. */
		if(stat(fname.monitor, &stat_buf))
		{
		    fprintf(stderr,
			"%s: Line %i: Warning: %s: No such file.\n",
			filename, lines_read, fname.monitor
		    );
		}
            }
            /* DBIn */
            else if(!strcasecmp(parm, "DBIn"))
            {
                if(fname.db_in_set == 0)
                {
                    strncpy(
                        fname.db_in,
                        val,
                        PATH_MAX + NAME_MAX
                    );
		    fname.db_in[PATH_MAX + NAME_MAX - 1] = '\0';
		    fname.db_in_set = 1;
		}
                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.db_in))
                {
                    sprintf(temp_name, "%s/%s",
                        dname.swserv_db_dir,
                        fname.db_in
                    );
		    strncpy(fname.db_in, temp_name, PATH_MAX + NAME_MAX);
                }
		fname.db_in[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Check if exists. */
                if(stat(fname.db_in, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such file.\n",
                        filename, lines_read, fname.db_in
                    );
                }
            }
            /* DBOut */ 
            else if(!strcasecmp(parm, "DBOut"))
            {
                if(fname.db_out_set == 0)
                {
                    strncpy(
                        fname.db_out,
                        val,
                        PATH_MAX + NAME_MAX
                    );
		    fname.db_out[PATH_MAX + NAME_MAX - 1] = '\0';
                    fname.db_out_set = 1;
                }
                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.db_out))
                {
                    sprintf(temp_name, "%s/%s",
                        dname.swserv_db_dir,
                        fname.db_out
                    );
                    strncpy(fname.db_out, temp_name, PATH_MAX + NAME_MAX);
                }
		fname.db_out[PATH_MAX + NAME_MAX - 1] = '\0';


/* Check if fname.db_out exists, if so, warn that it will be overwritten. */

            }
	    /* OCSFile */
            else if(!strcasecmp(parm, "OCSFile"))
            {
                if(fname.ocs_set == 0)
                {
                    strncpy(
                        fname.ocs,
                        val,
                        PATH_MAX + NAME_MAX
                    );
		    fname.ocs[PATH_MAX + NAME_MAX - 1] = '\0';
                    fname.ocs_set = 1;
                }
                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.ocs))
                {
                    sprintf(temp_name, "%s/%s",
                        dname.swserv_db_dir,
                        fname.ocs
                    );   
                    strncpy(fname.ocs, temp_name, PATH_MAX + NAME_MAX);
                }
                fname.ocs[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Check if exists. */
                if(stat(fname.ocs, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: File not found.\n",
                        filename, lines_read, fname.ocs
                    );
                }
            }
            /* OPMFile */
            else if(!strcasecmp(parm, "OPMFile"))
            {
                if(fname.opm_set == 0)
                {
                    strncpy(
                        fname.opm,
                        val,
                        PATH_MAX + NAME_MAX
                    );
		    fname.opm[PATH_MAX + NAME_MAX - 1] = '\0';
                    fname.opm_set = 1;
                }
		/* Set full path as needed. */
		if(!ISPATHABSOLUTE(fname.opm))
		{
		    sprintf(temp_name, "%s/%s",
			dname.swserv_db_dir,
			fname.opm
		    );
		    strncpy(fname.opm, temp_name, PATH_MAX + NAME_MAX);
		}
                fname.opm[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Check if exists. */
                if(stat(fname.opm, &stat_buf))
                {
                    fprintf(stderr,
                        "%s: Line %i: Warning: %s: No such file.\n",
                        filename, lines_read, fname.opm
                    );
                }
            }
            /* PrimaryLogFile */
            else if(!strcasecmp(parm, "PrimaryLogFile"))
            {
                strncpy(
                    fname.primary_log,
                    val,
                    PATH_MAX + NAME_MAX
                );
		fname.primary_log[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.primary_log))
                {
                    sprintf(temp_name, "%s/%s",
                        dname.swserv_logs_dir,
                        fname.primary_log
                    );
                    strncpy(fname.primary_log, temp_name, PATH_MAX + NAME_MAX);
                }
                fname.primary_log[PATH_MAX + NAME_MAX - 1] = '\0';
                continue;
            }
            /* ConListHTML */
            else if(!strcasecmp(parm, "ConListHTML"))
            {
                strncpy(
                    fname.conlist_html,
                    val,
                    PATH_MAX + NAME_MAX
                );
		fname.conlist_html[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.conlist_html))
                {
		    strncpy(
			fname.conlist_html,
			PrefixPaths(dname.public_html_dir, fname.conlist_html),
			PATH_MAX + NAME_MAX
		    );
                }
                fname.conlist_html[PATH_MAX + NAME_MAX - 1] = '\0';
            }
            /* ScoresHTML */   
            else if(!strcasecmp(parm, "ScoresHTML"))
            {
                strncpy(
                    fname.scores_html,
                    val,
                    PATH_MAX + NAME_MAX
                );
		fname.scores_html[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.scores_html))
                {
                    strncpy(
                        fname.scores_html,
                        PrefixPaths(dname.public_html_dir, fname.scores_html),
                        PATH_MAX + NAME_MAX
                    );
                }
                fname.scores_html[PATH_MAX + NAME_MAX - 1] = '\0';
            }
            /* EventsHTML */   
            else if(!strcasecmp(parm, "EventsHTML"))
            {
                strncpy(
                    fname.events_html,
                    val,
                    PATH_MAX + NAME_MAX
                );
		fname.events_html[PATH_MAX + NAME_MAX - 1] = '\0';

                /* Set full path as needed. */
                if(!ISPATHABSOLUTE(fname.events_html))
                {
                    strncpy(
                        fname.events_html,
                        PrefixPaths(dname.public_html_dir, fname.events_html), 
                        PATH_MAX + NAME_MAX
                    );
                }
                fname.events_html[PATH_MAX + NAME_MAX - 1] = '\0';
            }


	    /* ********************************************************* */
	    /* System parameters. */

            /* IntervalObjectValues */
            else if(!strcasecmp(parm, "IntervalObjectValues"))
            {
		sysparm.int_object_values = atol(val);
		if(sysparm.int_object_values < MIN_OBJECT_UPDATE_INT)
		{
		    fprintf(stderr,
   "%s: Line %i: Warning: IntervalObjectValues too low, setting to minimum %i.\n",
			filename,
			lines_read,
			MIN_OBJECT_UPDATE_INT
		    );
		    sysparm.int_object_values = MIN_OBJECT_UPDATE_INT;
		}

		next.object_values = cur_millitime + sysparm.int_object_values;
            }
            /* IntervalWeaponValues */
            else if(!strcasecmp(parm, "IntervalWeaponValues"))
            {
		sysparm.int_weapon_values = atol(val);
                if(sysparm.int_weapon_values < MIN_OBJECT_UPDATE_INT)
                {
                    fprintf(stderr,
   "%s: Line %i: Warning: IntervalWeaponValues too low, setting to minimum %i.\n",
                        filename,
                        lines_read,
			MIN_OBJECT_UPDATE_INT
                    );
                    sysparm.int_weapon_values = MIN_OBJECT_UPDATE_INT;
                }
                next.weapon_values = cur_millitime + sysparm.int_weapon_values;
            }
            /* IntervalAUXStats */
            else if(!strcasecmp(parm, "IntervalAUXStats"))
            {
                sysparm.int_aux_con_stats = atol(val);
                if(sysparm.int_aux_con_stats < 1000)
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: IntervalAUXStats too low, setting to minimum 1000.\n",
                        filename,
                        lines_read
                    );
                    sysparm.int_aux_con_stats = 1000;
                }
            }
            /* IntervalOSStats */
	    else if(!strcasecmp(parm, "IntervalOSStats"))
            {
                sysparm.int_os_stats = atol(val);	/* In seconds. */
                if(sysparm.int_os_stats < 1)
                {
                    fprintf(stderr,
   "%s: Line %i: Warning: IntervalOSStats too low, setting to minimum %i.\n",
                        filename,
                        lines_read,
			1
                    );
                    sysparm.int_os_stats = 1;
                }

		next.os_stats = cur_systime;	/* Next should be now. */
            }
            /* IntervalDBSave */
            else if(!strcasecmp(parm, "IntervalDBSave"))
            {
                sysparm.int_db_save = atol(val);	/* In seconds. */
                if(sysparm.int_db_save < 1)
                {
                    fprintf(stderr,
   "%s: Line %i: Warning: IntervalDBSave too low, setting to minimum %i.\n",
                        filename,
                        lines_read,
			1
                    );
                    sysparm.int_db_save = 1;
                }

                next.db_save = cur_systime + sysparm.int_db_save;
            }
            /* IntervalStatisticsExport */
            else if(!strcasecmp(parm, "IntervalStatisticsExport"))
            {
                sysparm.int_stats_export = atol(val);	/* In seconds. */
                if(sysparm.int_stats_export < 1)
                {
                    fprintf(stderr,
 "%s: Line %i: Warning: IntervalStatisticsExport too low, setting to minimum %i.\n",
                        filename,
                        lines_read,
			1
                    );
                    sysparm.int_stats_export = 1;
                }

                next.stats_export = cur_systime + sysparm.int_stats_export;
            }
            /* MaxConnections */
            else if(!strcasecmp(parm, "MaxConnections"))
            {
                sysparm.max_connections = atol(val);
                if(sysparm.max_connections < 1)
                {
                    fprintf(stderr,
           "%s: Line %i: Error: MaxConnections cannot be less than 1.\n",
                        filename,
                        lines_read
                    );
                    sysparm.max_connections = 1;
                }
            }
            /* MaxGuests */
            else if(!strcasecmp(parm, "MaxGuests"))
            {
                sysparm.max_guests = atol(val);
                if(sysparm.max_guests < 0)
                    sysparm.max_guests = 0;
            }
            /* MaxAUXStatsConnections */
            else if(!strcasecmp(parm, "MaxAUXStatsConnections"))
            {
                sysparm.max_aux_connections = atol(val);
		if(sysparm.max_aux_connections < 0)
		    sysparm.max_aux_connections = 0;
            }
	    /* LoginTimeout */
	    else if(!strcasecmp(parm, "LoginTimeout"))
	    {
		sysparm.login_timeout = atol(val);	/* Seconds. */
                if(sysparm.login_timeout < 1)
                    sysparm.login_timeout = 1;
	    }
	    /* MaxFailedLogins */
	    else if(!strcasecmp(parm, "MaxFailedLogins"))
            {
                sysparm.max_failed_logins = atoi(val);
                if(sysparm.max_failed_logins < 0)
                    sysparm.max_failed_logins = 0;
            }
	    /* GuestLoginName */
            else if(!strcasecmp(parm, "GuestLoginName"))
            {
                strncpy(sysparm.guest_login_name, val, XSW_OBJ_NAME_MAX);
		sysparm.guest_login_name[XSW_OBJ_NAME_MAX - 1] = '\0';
            }
	    /* AllowGuestLogins */
            else if(!strcasecmp(parm, "AllowGuestLogins"))
            {
		sysparm.allow_guest = StringIsYes(val);
            }
	    /* ConnectNotify */
	    else if(!strcasecmp(parm, "ConnectNotify"))
            {
                sysparm.con_notify = StringIsYes(val);
	    }
	    /* SingleConnection */
	    else if(!strcasecmp(parm, "SingleConnection"))
            {
                sysparm.single_connection = StringIsYes(val);
            }
            /* CeaseFire */
            else if(!strcasecmp(parm, "CeaseFire"))
            {
                sysparm.cease_fire = StringIsYes(val);
            }
            /* HidePlayers */
            else if(!strcasecmp(parm, "HidePlayers"))
            {
                sysparm.hide_players = StringIsYes(val);
            }
            /* HomesDestroyable */
            else if(!strcasecmp(parm, "HomesDestroyable"))
            {
                sysparm.homes_destroyable = StringIsYes(val);
            }

	    /* lost_found_owner is set in the universe file. */

	    /* LogGeneral */
            else if(!strcasecmp(parm, "LogGeneral"))
            {
		sysparm.log_general = StringIsYes(val);
            }
            /* LogEvents */
            else if(!strcasecmp(parm, "LogEvents"))
            {
                sysparm.log_events = StringIsYes(val);
            }
            /* LogNet */
            else if(!strcasecmp(parm, "LogNet"))
            {
                sysparm.log_net = StringIsYes(val);
            }
            /* LogErrors */
            else if(!strcasecmp(parm, "LogErrors"))
            {
                sysparm.log_errors = StringIsYes(val);
            }

	    /* MessageWrongLogin */
	    else if(!strcasecmp(parm, "MessageWrongLogin"))
	    {
		strncpy(
		    sysparm.mesg_wrong_login,
		    val,
		    CS_MESG_MAX
		);
		sysparm.mesg_wrong_login[CS_MESG_MAX - 1] = '\0';
	    }
            /* MessageNoGuests */
            else if(!strcasecmp(parm, "MessageNoGuests"))
            {
                strncpy(
                    sysparm.mesg_no_guests,
                    val,
                    CS_MESG_MAX
                );
                sysparm.mesg_no_guests[CS_MESG_MAX - 1] = '\0';
            }
	    /* MessageWelcome */
            else if(!strcasecmp(parm, "MessageWelcome"))
            {
                strncpy(
                    sysparm.mesg_welcome,
                    val,
                    CS_MESG_MAX
                );
                sysparm.mesg_welcome[CS_MESG_MAX - 1] = '\0';
            }
            /* MessageLeave */
            else if(!strcasecmp(parm, "MessageLeave"))
            {
                strncpy(
                    sysparm.mesg_leave,
                    val,
                    CS_MESG_MAX
                );
                sysparm.mesg_leave[CS_MESG_MAX - 1] = '\0';
            }
            /* ******************************************************** */
            /* BeginSiteBan */
            else if(!strcasecmp(parm, "BeginSiteBan"))
            {
		/* Allocate new site ban entry. */
		if(total_sitebans < 0)
		    total_sitebans = 0;

		n = total_sitebans;
		total_sitebans++;

		siteban = (siteban_struct **)realloc(
		    siteban,
		    total_sitebans * sizeof(siteban_struct *)
		);
		if(siteban == NULL)
		{
		    total_sitebans = 0;
		    continue;
		}

		siteban[n] = (siteban_struct *)calloc(
		    1,
		    sizeof(siteban_struct)
		);
		if(siteban[n] == NULL)
		{
		    total_sitebans--;
		    continue;
		}

		/* Get pointer to structure. */
		sb_ptr = siteban[n];


		while(1)
		{
                    /* Free previous line and allocate/read next line. */
                    free(strptr); strptr = NULL;
                    strptr = FReadNextLineAlloc(fp, UNIXCFG_COMMENT_CHAR);
                    if(strptr == NULL) break;
                    lines_read++;

                    /* Fetch parameter. */
                    strptr2 = StringCfgParseParm(strptr);
                    if(strptr2 == NULL) continue;
                    strncpy(parm, strptr2, CFG_PARAMETER_MAX);
                    parm[CFG_PARAMETER_MAX - 1] = '\0';

                    /* Fetch value. */
                    strptr2 = StringCfgParseValue(strptr);
                    if(strptr2 == NULL) strptr2 = "0";
                    strncpy(val, strptr2, CFG_VALUE_MAX);
                    val[CFG_VALUE_MAX - 1] = '\0';


		    /* Address */
                    if(!strcasecmp(parm, "Address"))
                    {
			StringParseIP(
			    val,
			    &(sb_ptr->ip.part_u8[0]),
                            &(sb_ptr->ip.part_u8[1]),
                            &(sb_ptr->ip.part_u8[2]),
                            &(sb_ptr->ip.part_u8[3])
			);
		    }
                    /* Restrict */
                    else if(!strcasecmp(parm, "Restrict"))
                    {
			sb_ptr->restrict = atoi(val);
                    }
                    /* EndSiteBan */
                    else if(!strcasecmp(parm, "EndSiteBan"))
                    {
			break;
		    }
		}
	    }

	    /* ******************************************************** */

 	    /* Unknown parameter. */
	    else
	    {
		fprintf(stderr, "%s: Line %i: Unknown parameter: `%s'\n",
		    filename,
		    lines_read,
		    parm
		);
	    }
	}


	/* Close file. */
	fclose(fp); fp = NULL;


	return(0);
}
